import sys
import pytest
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_VIDEO
from sdl2 import blendmode


@pytest.mark.skip("not implemented")
def test_SDL_ComposeCustomBlendMode(self, with_sdl):
    pass
